(* Mathematica package *)

(*******************************************************)
(*****************************************Begin Package*)
BeginPackage["OrientationScores`GroupConvolutions`"]
Unprotect["OrientationScores`GroupConvolutions`*"]


GaussianKernelSE2::usage = "GaussianKernelSE2"
GaussianKernelR3::usage = "GaussianKernelR3[s\[Theta], {t\[Theta], t\[Xi], t\[Eta]}] constructs a 3D gaussian kernel based on the scales {t\[Theta], t\[Xi], t\[Eta]}.

GaussianKernelR3[s\[Theta], {t\[Theta], t\[Xi], t\[Eta]}, n\[Theta]] does the same but with the kernel size in \[Theta] direction restricted to 2*n\[Theta]+1.

For options see documentation."
GConvolve::usage = "GConvolve[kernel, os] applies SE(2) convolution with kernel on the orientation score. 

For options see documentation."
GaussianBlurSE2::usage = "GaussianBlurSE2[os, {t\[Theta], t\[Xi], t\[Eta]}] applies Gaussian blurring to the orientation score os, by means of SE(2) group 
convolution with a 3D gaussian kernel.

GaussianBlurSE2[os, {t\[Theta], t\[Xi], t\[Eta]},  n\[Theta]] does the same but with the Gaussian kernel size in \[Theta] direction restricted 
to 2*n\[Theta]+1.

For options see documentation."

(*******************************************************)
(*****************************************Begin Private*)
Begin["Private`"]

Options[GaussianKernelSE2] = {"CoordinateSystem" -> "Matrix"}
GaussianKernelSE2[s\[Theta]_, \[Sigma]List_, n\[Theta]_, OptionsPattern[]] := GaussianKernelSE2Code[s\[Theta], \[Sigma]List, n\[Theta],  OptionValue["CoordinateSystem"] == "Matrix"]
  
GaussianKernelSE2Code =  Compile[{{s\[Theta], _Real}, {\[Sigma]List, _Real, 1}, {n\[Theta], _Integer},{matrixCoordinatesQ,True | False}},
	Module[{\[Sigma]\[Theta], \[Sigma]\[Xi], \[Sigma]\[Eta], \[Sigma]\[Xi]\[Eta], n\[Theta]Max, e\[Xi], e\[Eta], \[Xi], \[Eta]},
		{\[Sigma]\[Theta], \[Sigma]\[Xi], \[Sigma]\[Eta]} = \[Sigma]List;
		\[Sigma]\[Xi]\[Eta] = Max[{10, \[Sigma]\[Xi], \[Sigma]\[Eta]}];
		n\[Theta]Max = Min[Round[\[Pi]/(2 s\[Theta])], n\[Theta]];
		
		Table[
			\[Theta]2 = If[matrixCoordinatesQ,\[Theta],-\[Theta]+\[Pi]/2];
			e\[Xi] = {Cos[\[Theta]2], Sin[\[Theta]2]};
			e\[Eta] = {-Sin[\[Theta]2], Cos[\[Theta]2]};
			Table[
				\[Xi] = {x, y}.e\[Xi];
				\[Eta] = {x, y}.e\[Eta];
				
				E^(-0.5(\[Xi]^2/\[Sigma]\[Xi]^2) -0.5( \[Eta]^2/\[Sigma]\[Eta]^2 ) - 0.5( \[Theta]^2/\[Sigma]\[Theta]^2 ) )/(\[Sigma]\[Theta]*\[Sigma]\[Eta]*\[Sigma]\[Xi]*(2 \[Pi])^(3/2))
				, {x, -4*\[Sigma]\[Xi]\[Eta], 4*\[Sigma]\[Xi]\[Eta], 1}, {y, -4*\[Sigma]\[Xi]\[Eta], 4*\[Sigma]\[Xi]\[Eta], 1}]
			, {\[Theta], -s\[Theta]*n\[Theta]Max, s\[Theta]*n\[Theta]Max, s\[Theta]}]
		]
	]
      
Options[GaussianKernelR3] = {"CoordinateSystem" -> "Matrix"}
GaussianKernelR3[s\[Theta]_, {t\[Theta]_, t\[Xi]_, t\[Eta]_}, n\[Theta]_:\[Infinity], OptionsPattern[]] :=
	Module[{\[Sigma]\[Theta], \[Sigma]\[Xi], \[Sigma]\[Eta], n\[Theta]Max, n\[Xi]\[Eta], kernel},
		{\[Sigma]\[Theta], \[Sigma]\[Xi], \[Sigma]\[Eta]} = (Sqrt[2 #]) & /@ {t\[Theta], t\[Xi], t\[Eta]};
		\[Sigma]\[Theta] = \[Sigma]\[Theta]/s\[Theta];
		n\[Theta]Max = Min[Round[4 \[Sigma]\[Theta]], n\[Theta], Round[\[Pi]/(2 s\[Theta])]];
		n\[Xi]\[Eta] = Max[Round[4*{\[Sigma]\[Xi], \[Sigma]\[Eta]}]];
		kernel = ConstantArray[ 0, {2 n\[Theta]Max, 2 n\[Xi]\[Eta], 2 n\[Xi]\[Eta]} + 1];
		kernel[[n\[Theta]Max + 1, n\[Xi]\[Eta] + 1, n\[Xi]\[Eta] + 1]] = 1;
		kernel = GaussianFilter[kernel, {Dimensions[kernel]/2, {\[Sigma]\[Theta], \[Sigma]\[Xi], \[Sigma]\[Eta]}}];
		If[OptionValue["CoordinateSystem"]!="Matrix",
			kernel=Transpose/@kernel
		];
		Return[kernel]
  ]


Options[GConvolve] = {"Periodicity"->\[Pi],"CoordinateSystem" -> "Matrix"}
GConvolve[gkernel_, os_, OptionsPattern[]] := 
	Module[{dim\[Theta] = 1, dimx = 1, dimy = 1, c = 1, periodicity,rotateDirection,gkernelIm},
		periodicity = OptionValue["Periodicity"];
		{dim\[Theta], dimx, dimy} = Dimensions[gkernel];
		c = Ceiling[dim\[Theta]/2];
		rotateDirection = If[OptionValue["CoordinateSystem"] == "Matrix", 1, -1];
		gkernelIm=Image/@gkernel;
		Table[
			Total@Table[
				ListConvolve[
					ImageData@ImageRotate[gkernelIm[[c + i]], rotateDirection*(\[Theta]ii - 1)*periodicity/Length[os], Full]
					,
					os[[Mod[\[Theta]ii + i, Length[os], 1]]]
					,
					{-Ceiling[{dimx, dimy}/2], Ceiling[{dimx, dimy}/2]}]
					, {i, -(c - 1), (c - 1), 1}]
			, {\[Theta]ii, 1, Length[os], 1}]
		];


Options[GaussianBlurSE2] = {"Periodicity" -> \[Pi], "CoordinateSystem" -> "Matrix"}
GaussianBlurSE2[os_, {t\[Theta]_, t\[Xi]_, t\[Eta]_}, n\[Theta]_:\[Infinity], OptionsPattern[]] := 
	Module[{s\[Theta], gkernel},
		s\[Theta] = OptionValue["Periodicity"]/Length[os];
		gkernel = GaussianKernelR3[s\[Theta], {t\[Theta], t\[Xi], t\[Eta]}, n\[Theta], "CoordinateSystem" -> OptionValue["CoordinateSystem"]];
		Return[GConvolve[Reverse@gkernel, os, "Periodicity" -> OptionValue["Periodicity"], "CoordinateSystem" -> OptionValue["CoordinateSystem"]]]
		]
  
		
(*******************************************************)
(*******************************************End Private*)
End[]

(*******************************************************)
(*******************************************End Package*)
Protect["OrientationScores`GroupConvolutions`*"]
EndPackage[]