(* Mathematica package *)

(*******************************************************)
(*****************************************Begin Package*)
BeginPackage["OrientationScores`Transformation`"]
Unprotect["OrientationScores`Transformation`*"]



CakeWaveletStack::usage = "CakeWaveletStack[size,s\[CurlyPhi]] returns the set 
	of rotated wavelets with standard paramters (see Options[CakeWaveletStack]).
	The kernel stack has dimensions No x size x size, where No = \!\(\*FractionBox[\(s\[CurlyPhi]\), \(periodicity\)]\). 
	By default periodicity = \[Pi], meaning that the wavelets are constructed 
	from 0 to \[Pi]. Parameter s\[CurlyPhi] specifies the angular resolution: 
	the number of kernels per \[Pi] radians. 

Other parameters can be adjusted through the options or by calling the function 
	as follows:
 
CakeWaveletStack[size,s\[CurlyPhi],kSpline,q,periodicity,\[Rho]Inflection,kMn,\[Sigma]s,matrixCoordinatesQ]."

GaborWaveletStack::usage = "GaborWaveletStack[a,s\[CurlyPhi]] returns the set 
	of rotated Gabor wavelets at scale a, with standard paramters (see Options[GaborWaveletStack]).
	The kernel stack has dimensions No x size x size, where No = \!\(\*FractionBox[\(s\[CurlyPhi]\), \(periodicity\)]\),
	and where size = 8*a*Sqrt[\[Epsilon]]. By default periodicity = \[Pi], meaning that 
	the wavelets are constructed from 0 to \[Pi]. Parameter s\[CurlyPhi] specifies 
	the angular resolution: the number of kernels per \[Pi] radians."


WaveletTransform2D::usage = "WaveletTransform2D[kernelStack,im] correlates each 
	kernel with the image. If the kernel dimensions and image dimensions 
	coincide, correlation is done using (conjugate) Fourier multiplication, else
	 it is done using correlation."

OrientationScoreTransform::usage = "OrientationScoreTransform[image, s\[CurlyPhi]] 
	constructs an orientation score from the image using cake wavelets with 
	default paramters (see Options[OrientationScoreTransform] and 
	?CakeWaveletStack). By default this is done through Fourier domain 
	multiplication, unless it is specified (by using the options) otherwise."


(*******************************************************)
(*****************************************Begin Private*)
Begin["Private`"]






(***********************************)
(******************CakeWaveletStack Private Functions*)

(*PolarCoordinateGridRadial returns a matrix in which each element gives the corresponding radial coordinate (with the origin in the center of the matrix*)
PolarCoordinateGridRadial=Compile[{{size,_Integer}},
	Table[
		Sqrt[x^2+y^2]
		,{x,-((size-1)/2)-Mod[(size-1)/2,1],(size-1)/2-Mod[(size-1)/2,1]}
		,{y,-((size-1)/2)-Mod[(size-1)/2,1],(size-1)/2-Mod[(size-1)/2,1]}]
];
(*PolarCoordinateGridRadial returns a matrix in which each element gives the corresponding radial coordinate (with the origin in the center of the matrix*)
PolarCoordinateGridAngular=Compile[{{size,_Integer}},
	Table[
		Arg[x+I y]
		,{x,-((size-1)/2)-Mod[(size-1)/2,1],(size-1)/2-Mod[(size-1)/2,1]}
		,{y,-((size-1)/2)-Mod[(size-1)/2,1],(size-1)/2-Mod[(size-1)/2,1]}]
];
(*MnWindow gives the radial windowing matrix for sampling the fourier domain*)
MnWindow=Compile[{{size,_Integer},{n,_Integer},{\[Rho]inflection,_Real}},
	Module[{\[Rho]Matrix},
		\[Rho]Matrix=$MachineEpsilon+1/Sqrt[(2*\[Rho]inflection^2)/(1+2 n)] PolarCoordinateGridRadial[size];
		Total[Table[E^-\[Rho]Matrix^2*\[Rho]Matrix^(2k)/k!,{k,0,n}]]
	]
,{{k!,_Real}}
];
(*BSplineScalarFunc is a compile spline basis function of arbitrary order. The function is defined for scalars, list and 2D matrix objects*)
BSplineScalarFunc=Compile[{{n,_Integer},{x,_Real}},
	Module[{function,intervalCheck},
		Total[
			Table[
				function=1/(2 (n+1-1)!) Total[Table[Binomial[n+1,k](x+(n+1)/2-k)^(n+1-1) (-1)^k Sign[i+(n+1)/2-k],{k,0,n+1}]];
				intervalCheck=Round[If[i<(n+1-1)/2,UnitStep[x-(i-1/2+$MachineEpsilon)]*UnitStep[-(x-(i+1/2))],UnitStep[x-(i-1/2+$MachineEpsilon)]*UnitStep[-(x-(i+1/2-$MachineEpsilon))]]];
				function*intervalCheck
				,{i,(1-n-1)/2,(n-1+1)/2}
			]
		]
	]
];
BSplineListFunc=Compile[{{n,_Integer},{x,_Real,1}},
	Module[{function,intervalCheck},
		Total[
			Table[
				function=1/(2 (n+1-1)!) Total[Table[Binomial[n+1,k](x+(n+1)/2-k)^(n+1-1) (-1)^k Sign[i+(n+1)/2-k],{k,0,n+1}]];
				intervalCheck=Round[If[i<(n+1-1)/2,UnitStep[x-(i-1/2+$MachineEpsilon)]*UnitStep[-(x-(i+1/2))],UnitStep[x-(i-1/2+$MachineEpsilon)]*UnitStep[-(x-(i+1/2-$MachineEpsilon))]]];
				function*intervalCheck
				,{i,(1-n-1)/2,(n-1+1)/2}
			]
		]
	]
];
BSplineMatrixFunc=Compile[{{n,_Integer},{x,_Real,2}},
	Module[{function,intervalCheck},
		Total[
			Table[
				function=1/(2 (n+1-1)!) Total[Table[Binomial[n+1,k](x+(n+1)/2-k)^(n+1-1) (-1)^k Sign[i+(n+1)/2-k],{k,0,n+1}]];
				intervalCheck=Round[If[i<(n+1-1)/2,UnitStep[x-(i-1/2+$MachineEpsilon)]*UnitStep[-(x-(i+1/2))],UnitStep[x-(i-1/2+$MachineEpsilon)]*UnitStep[-(x-(i+1/2-$MachineEpsilon))]]];
				function*intervalCheck
				,{i,(1-n-1)/2,(n-1+1)/2}
			]
		]
	]
];
(*WindowGauss retuns the spatial Gauss envelope*)
WindowGauss=Compile[{{size,_Integer},{\[Sigma]s,_Real}},
	Table[
		E^(-(x^2/(2\[Sigma]s^2))-y^2/(2\[Sigma]s^2))
		,{x,-((size-1)/2)-Mod[(size-1)/2,1],(size-1)/2-Mod[(size-1)/2,1]}
		,{y,-((size-1)/2)-Mod[(size-1)/2,1],(size-1)/2-Mod[(size-1)/2,1]}
	]
];
(*CakeWaveletStackFourier constructs the cake wavelets in the Fourier domain (note that windowing in the spatial domain is still required after this*)
CakeWaveletStackFourier=Compile[{{size,_Integer},{s\[CurlyPhi],_Real},{kspline,_Integer},{q,_Integer},{\[Rho]Inflection,_Real},{kMn,_Integer},{matrixCoordinatesQ,True|False}},
	Module[{mnWindow,angleGrid},
		mnWindow=MnWindow[size,kMn,\[Rho]Inflection];
		angleGrid=PolarCoordinateGridAngular[size];
		Table[
			mnWindow*BSplineMatrixFunc[kspline,If[matrixCoordinatesQ,Mod[angleGrid-\[Theta]-\[Pi]/2,2\[Pi],-\[Pi]]/s\[CurlyPhi],Mod[angleGrid+\[Theta]+\[Pi],2\[Pi],-\[Pi]]/s\[CurlyPhi]]]
			,{\[Theta],\[Pi],2\[Pi]-s\[CurlyPhi]/q,(s\[CurlyPhi]/q)}
		]
	]
]
(*ErfSet retuns a set of 2D error functions. This function is used to cut the wavelets in two (in the spatial domain)*)
ErfSet=Compile[{{size,_Real},{No,_Integer}},
	Module[{d\[Theta]},
		Table[
			Table[
				1/2 (1+Erf[x Cos[(2 (-1+i) \[Pi])/No]+y Sin[(2 (-1+i) \[Pi])/No]])
				,{x,-((size-1)/2)-Mod[(size-1)/2,1],(size-1)/2-Mod[(size-1)/2,1]}
				,{y,-((size-1)/2)-Mod[(size-1)/2,1],(size-1)/2-Mod[(size-1)/2,1]}
			]
			,{i,1,No}
		]
	]
];








(***********************************)
(******************CakeWaveletStack*)

Options[CakeWaveletStack]={
"SplineOrder"->3,
"AngularSubSampleRate"->1,
"Periodicity"->\[Pi],
"InflectionPoint"->"Automatic",
"MnOrder"->8,
"GaussianWindowSigma"-> "Automatic",
"InflectionPoint"->"Automatic",
"CoordinateSystem"->"Matrix",
"SingleVsDoubleSided"->"Double"};
(*Function definition for direct options input*)
CakeWaveletStack[size_,s\[CurlyPhi]_,kSpline_,q_,periodicity_,\[Rho]Inflection_,kMn_,\[Sigma]s_,matrixCoordinatesQ_,singleOrDouble_]:=
	Module[{cakeF,cakeIF,window,cake},
		cakeF=CakeWaveletStackFourier[size,s\[CurlyPhi],kSpline,q,\[Rho]Inflection,kMn,matrixCoordinatesQ];
		(*cakeF[[All,Ceiling[(size-1)/2]+2,Ceiling[(size-1)/2]+2]]=0*s\[CurlyPhi]/periodicity;(*The central frequency overlaps with all rotated kernels, it should still ad up to 1*)*)
		cakeF[[All,Ceiling[(size-1)/2]+1-1;;Ceiling[(size-1)/2]+1+1,Ceiling[(size-1)/2]+1-1;;Ceiling[(size-1)/2]+1+1]]=s\[CurlyPhi]/periodicity;(*The central frequency overlaps with all rotated kernels, it should still ad up to 1*)
		cakeIF=RotateRight[InverseFourier[RotateLeft[#,Floor[{size,size}/2]]],Floor[{size,size}/2]]&/@cakeF;
		window=WindowGauss[size,\[Sigma]s];
		cake=window*#&/@cakeIF;
		If[periodicity==2\[Pi],
			cake=Flatten[{cake,Conjugate[Reverse[Reverse[#]]&/@cake]},1];
		];
		If[singleOrDouble=="Single",
			cake=cake*ErfSet[size,periodicity/s\[CurlyPhi]];
		];
		Return[cake]
	]
(*Function definition that uses the OptionsPattern*)
CakeWaveletStack[size_,s\[CurlyPhi]_,OptionsPattern[]]:=
	Module[{kSpline,q,periodicity,\[Rho]Inflection,kMn,\[Sigma]s,matrixCoordinatesQ,cakeStack,singleOrDouble},
		(****************************)
		(***********Check Parameters*)
		kSpline=OptionValue["SplineOrder" ];
		q=OptionValue["AngularSubSampleRate"];
		periodicity=OptionValue["Periodicity"];
		Switch[OptionValue["InflectionPoint"]
			,"Automatic",
			\[Rho]Inflection=(size-1)/2 0.8;
			,_,
			\[Rho]Inflection=OptionValue["InflectionPoint"]
		];
		kMn=OptionValue["MnOrder"];
		Switch[OptionValue["GaussianWindowSigma"]
			,"Automatic",
			\[Sigma]s=(size-1)/4;
			,_,
			\[Sigma]s=OptionValue["GaussianWindowSigma"]
		];
		If[OptionValue["CoordinateSystem"]=="Matrix",matrixCoordinatesQ=True,matrixCoordinatesQ=False];
		singleOrDouble=OptionValue["SingleVsDoubleSided"];
		
		(***************************)
		(***********Construct Stack*)
		cakeStack=CakeWaveletStack[size,s\[CurlyPhi],kSpline,q,periodicity,\[Rho]Inflection,kMn,\[Sigma]s,matrixCoordinatesQ,singleOrDouble];
		Return[cakeStack]
	]








(***********************************)
(******************GaborWaveletStack*)

Options[GaborWaveletStack] = {
  "\[Epsilon]" -> 4,
  "KernelSize" -> "Automatic",
  "k0" -> {0, 3},
  "Periodicity" -> \[Pi],
  "SingleVsDoubleSided" -> "Double",
  "CoordinateSystem" -> "Matrix"
  };
GaborWaveletStack[a_, s\[CurlyPhi]_, OptionsPattern[]] :=
Module[{\[Epsilon], A, k0, size, periodicity, stack, singleOrDouble, matrixCoordinatesQ},
	(*Parameters:*)
	\[Epsilon] = OptionValue["\[Epsilon]"];
	A = ( {
			{\[Epsilon]^(-1/2), 0},
			{0, 1}
			} );
	k0 = OptionValue["k0"];
	Switch[OptionValue["KernelSize"]
		,"Automatic",
		size = Floor[8*Sqrt[\[Epsilon]]*a]
		,_Integer,
		size = OptionValue["KernelSize"]
		,_,
		size = Floor[8*Sqrt[\[Epsilon]]*a]
	];
	If[EvenQ[size], size = size + 1];
	periodicity = OptionValue["Periodicity"];
	singleOrDouble = OptionValue["SingleVsDoubleSided"];
	matrixCoordinatesQ = OptionValue["CoordinateSystem"] == "Matrix";
	(*The stack:*)
	stack = GenerateKernelSetGabor[A, k0, a, Round[\[Pi]/s\[CurlyPhi]], size, matrixCoordinatesQ];
	If[periodicity == 2 \[Pi],
		stack = Flatten[{stack, Conjugate[Reverse[Reverse[#]] & /@ stack]}, 1];];
	If[singleOrDouble == "Single",
		stack = stack*ErfSet[size, periodicity/s\[CurlyPhi]];];
    Return[stack]
];
GenerateKernelSetGabor=Compile[{{A,_Real,2},{k0,_Real,1},{a,_Real},{n\[Theta],_Real},{dim,_Real},{matQ,True|False}},
	Module[{d\[Theta]},
		d\[Theta]=(\[Pi])/n\[Theta];
		0.00000000001+(1/((4 a \[Pi])/E^(9/2)))(1/a)*Table[
			Table[
				With[{xx={Cos[-\[Theta]+\[Pi]/2], -Sin[-\[Theta]+\[Pi]/2]}.{x,y},yy={Sin[-\[Theta]+\[Pi]/2], Cos[-\[Theta]+\[Pi]/2]}.{x,y}},
				 Exp[I (k0.({yy,xx}/a))]Exp[-(1/2) Norm[(A.({yy,xx}/a))]^2]]
			,{x,-Floor[dim/2],Floor[dim/2]-(1-2*Mod[dim/2,1])}
			,{y,-Floor[dim/2],Floor[dim/2]-(1-2*Mod[dim/2,1])}]
		,{\[Theta],If[matQ,\[Pi],\[Pi]+\[Pi]/2],If[matQ,2*\[Pi]-d\[Theta],+\[Pi]/2+d\[Theta]],If[matQ,d\[Theta],-d\[Theta]]}]
	]
];










(***********************************)
(****************WaveletTransform2D*)

WaveletTransform2D[kernels_,im_,fourierQ_:1]:=
	Module[{imF,kernelsF,os,dimkernelx1,dimkernely1,dimkernelx2,dimkernely2},
		If[Dimensions[kernels][[-2;;-1]]==Dimensions[im]&&fourierQ==1,
			(*Using the Fourier transform*)
			imF=Fourier[im];
			kernelsF=Fourier[Reverse[#,{1,2}]]&/@kernels;
			os=RotateRight[InverseFourier[#*imF],Ceiling[0.1+Dimensions[im]/2]]&/@kernelsF;
		,
			(*Using Correlation*)
			{dimkernelx1,dimkernely1}=Dimensions[kernels[[1]]]/2//Ceiling;
			{dimkernelx2,dimkernely2}=Dimensions[kernels[[1]]];
			If[EvenQ[dimkernelx2],dimkernelx2=dimkernelx2/2+1,dimkernelx2=dimkernelx2/2//Ceiling];
			If[EvenQ[dimkernely2],dimkernely2=dimkernely2/2+1,dimkernely2=dimkernely2/2//Ceiling];
			os=Table[
				ListCorrelate[kernels[[i]],im,{{-dimkernelx1,-dimkernely1},{dimkernelx2,dimkernely2}}]
				,{i,1,Length[kernels],1}
				]
		];
		Return[os]
	]












(***********************************)
(****************WaveletTransform2D*)

Options[OrientationScoreTransform]={
"WaveletSize"->151,
"SplineOrder"->3,
"AngularSubSampleRate"->1,
"Periodicity"->\[Pi],
"InflectionPoint"->"Automatic",
"MnOrder"->8,
"GaussianWindowSigma"-> "Automatic",
"InflectionPoint"->"Automatic",
"CoordinateSystem"->"Matrix",
"SingleVsDoubleSided"->"Double",
"Method"->"SpatialConvolution"};

OrientationScoreTransform::dim = "Warning! Method \"Fourier\" can not be used if the image dimensions are not equal. Convolution based method with kernel size `1`x`1` used instead.";
OrientationScoreTransform::size = "Warning! Wavelet size can not be specified with method \"Fourier\", size is set equal to the image size.";

OrientationScoreTransform[im_,s\[CurlyPhi]_,OptionsPattern[]]:=
Module[{dimx,dimy,size,kSpline,q,periodicity,\[Rho]Inflection,kMn,\[Sigma]s,matrixCoordinatesQ,cakeStack,os,singleOrDouble},
	(****************************)
	(***********Check Parameters*)
	{dimx,dimy}=Dimensions[im][[1;;2]];
	If[OptionValue["Method"]=="Fourier",
		If[dimx!=dimy,
			Message[OrientationScoreTransform::dim,OptionValue["WaveletSize"]];
			size=OptionValue["WaveletSize"];
			,
			If[OptionValue["WaveletSize"]!=Options[OrientationScoreTransform,"WaveletSize"][[1,2]],
				Message[OrientationScoreTransform::size];
			];
				size=dimx;
		]
		,
		size=OptionValue["WaveletSize"]
	];
	kSpline=OptionValue["SplineOrder" ];
	q=OptionValue["AngularSubSampleRate"];
	periodicity=OptionValue["Periodicity"];
	Switch[OptionValue["InflectionPoint"]
		,"Automatic",
		\[Rho]Inflection=(size-1)/2 0.8;
		,_,
		\[Rho]Inflection=OptionValue["InflectionPoint"]
	];
	kMn=OptionValue["MnOrder"];
	Switch[OptionValue["GaussianWindowSigma"]
		,"Automatic",
		\[Sigma]s=(size-1)/4;
		,_,
		\[Sigma]s=OptionValue["GaussianWindowSigma"]
	];
	If[OptionValue["CoordinateSystem"]=="Matrix",matrixCoordinatesQ=True,matrixCoordinatesQ=False];
	singleOrDouble=OptionValue["SingleVsDoubleSided"];
	(***************************)
	(***********Construct Stack*)
	cakeStack=CakeWaveletStack[size,s\[CurlyPhi],kSpline,q,periodicity,\[Rho]Inflection,kMn,\[Sigma]s,matrixCoordinatesQ,singleOrDouble];
	os=WaveletTransform2D[cakeStack,im,If[OptionValue["Method"]=="Fourier",1,0]];
	Return[os]
]
















(*******************************************************)
(*******************************************End Private*)
End[]

(*******************************************************)
(*******************************************End Package*)
Protect["OrientationScores`Transformation`*"]
EndPackage[]