(* Mathematica package *)

(*******************************************************)
(*****************************************Begin Package*)
BeginPackage["RetinalImageAnalysis`Preprocessing`"]
Unprotect["RetinalImageAnalysis`Preprocessing`*"]

CreateApertureMask::usage = "CreateApertureMask[\!\(\*
StyleBox[\"im\",\nFontSlant->\"Italic\"]\)] automatically creates an binary mask of the field of view (aperture) of the fundus image.\n
CreateApertureMask[\!\(\*
StyleBox[\"im\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"tval\",\nFontSlant->\"Italic\"]\)] creates a mask image by thresholding with value \!\(\*
StyleBox[\"tval\",\nFontSlant->\"Italic\"]\).
\n\t \!\(\*
StyleBox[\"tval\",\nFontSlant->\"Italic\"]\) \[Equal] -1: The threshold value is automatically determined with the Mathematica function FindThreshold. This is the default option.
\n\t \!\(\*
StyleBox[\"tval\",\nFontSlant->\"Italic\"]\) \[Equal] -2: The function RegionBinarize is used with distance parameter d = 0.05, on a rescaled image.
\n\t \!\(\*
StyleBox[\"Else\",\nFontSlant->\"Italic\"]\): The specified value for \!\(\*
StyleBox[\"tval\",\nFontSlant->\"Italic\"]\) is used to threshold the image."



InpaintBackground::usage = "InpaintBackground[\!\(\*
StyleBox[\"im\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"mask\",\nFontSlant->\"Italic\"]\)] inpaints the image background (denoted by 0 in the mask), the function is based on the Mathematica function Inpaint.\n
InpaintBackground[\!\(\*
StyleBox[\"im\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"mask\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"fgErosionSize\",\nFontSlant->\"Italic\"]\), \!\(\*
StyleBox[\"speedUpFactor\",\nFontSlant->\"Italic\"]\)] inpaints the background. Using \!\(\*
StyleBox[\"fgErosionSize\",\nFontSlant->\"Italic\"]\) the mask foreground is eroded, the default value is 2. The parameter \!\(\*
StyleBox[\"speedUpFactor\",\nFontSlant->\"Italic\"]\) decreases the calculation time by down-sampling the image with a factor \!\(\*
StyleBox[\"speedUpFacor\",\nFontSlant->\"Italic\"]\). Even with this speed up factor, the image data within the foreground stays in tact."




HighPass::usage = "HighPass[im, \[Sigma]] high-pass filters an image by
removing the low frequency components in the image. The low frequency components
are estimated by Gaussian blurring the image with scale \[Sigma]."


ESPaintBackground::usage = "ESPaintBackground[im] inpaints the region outside 
the field of view.

ESPaintBackground[im,verticalBuffer] also replaces a number (verticalBuffer) of
rows on the top and bottom of the image."





IECorrection::usage = "IECorrection[im] applies illumination equalization
correction to the input image. The blurring standard deviation is determined 
automatically.

IECorrection[im,\[Sigma]] does the same but with use specified blurring standard
deviation."



ESPreprocessImage::usage = "ESPreprocessImage[im] preprocesses the image for further
automated image analysis. Preprocessing involves: background inpainting,
illumination correction and high-pass filtering."

PreprocessImage::usage = "PreprocessImage[im] preprocesses the image for further
automated image analysis. Preprocessing involves: background inpainting,
illumination correction and high-pass filtering."

(*******************************************************)
(*****************************************Begin Private*)
Begin["Private`"]





HighPass[im_, \[Sigma]_] := im - ImageData@GaussianFilter[Image[im], {3 \[Sigma], \[Sigma]}];
HighPass[im_, \[Sigma]_,fast_] := im - HeavyBlurringFast[im,\[Sigma]];







BackgroundMirror = Compile[{{im, _Real, 2}, {bgPos, _Integer, 2}, {bgPosMirrored, _Integer, 2}},
	Module[{imNew, x, y, x1, y1},
		imNew = im;
		Do[
			{x, y} = bgPos[[pointNr]];
			{x1, y1} = bgPosMirrored[[pointNr]];
			imNew[[x, y]] = im[[x1, y1]]
		, {pointNr, 1, Length[bgPos]}];
		imNew
	]
]






ESPaintBackground[im_, verticalBuffer_:15, mirrorMapIn_: 0] := 
 Module[{mirrorMap, imNew, applicationPaths, applicationPath},
  If[mirrorMapIn == 0,
   applicationPaths = Extract[$Path, Position[
      Length /@ (StringCases[#, "Applications"] & /@ $Path)
      , 1]
     ];
   applicationPath = First@Extract[
      applicationPaths,
      Position[
       Length /@ (FileNames[# <> "\\RetinalImageAnalysis"] & /@ 
          applicationPaths)
       , 1]
      ];
   
   If[Not[MatchQ[Dimensions@ESGlobal`mirrorMap, {2, _, 2}]], 
    ESGlobal`mirrorMap = 
     Import[applicationPath <> 
        "\\RetinalImageAnalysis\\Data\\EasyScanData\\mirrorMap.dat", 
       "String"] // ToExpression];
   mirrorMap = ESGlobal`mirrorMap;
   ,
   mirrorMap = mirrorMapIn];
  
  imNew = im;
  imNew[[1 ;; verticalBuffer]] = 
   Reverse@imNew[[verticalBuffer + 1 ;; 2*verticalBuffer]];
  imNew[[1024 + 1 - verticalBuffer ;; -1]] = 
   Reverse@
    imNew[[1024 + 1 - 2*verticalBuffer ;; 1024 - verticalBuffer]];
  imNew = BackgroundMirror[imNew, mirrorMap[[1]], mirrorMap[[2]]];
  imNew
  ]
  
  
  
  
  





InpaintBackground[im_,mask_,dilation_:2,speedUpFactor_:4]:=Module[{image,bgmask,fgmask,dimx,dimy,inpainted},
image=Image[im,Interleaving->False];
bgmask=Dilation[ColorNegate[Image[mask,"Bit",Interleaving->False]],dilation,Padding->0];
fgmask=ColorNegate[bgmask];
{dimx,dimy}=Dimensions[im];
inpainted=Inpaint[ImageResize[image,{dimy,dimx}/speedUpFactor//Round],ImageResize[bgmask,{dimy,dimx}/speedUpFactor//Round]];
ImageData[ImageAdd[ImageMultiply[fgmask,image],ImageMultiply[bgmask,ImageResize[inpainted,{dimy,dimx}]]],Interleaving->False]
]





LMax[list_] := Module[{lmax},
  lmax = {};
  Do[If[list[[i - 1]] < list[[i]] > list[[i + 1]], 
    AppendTo[lmax, i]], {i, 2, Length[list] - 1, 1}];
  lmax
  ]
CreateApertureMask[im_,thresholdIn_:-1]:=Module[{ds,threshold,histo,lmin,mask},
Switch[thresholdIn
	,-3,
	ds=Ceiling[Length[im]/256];
	histo=HistogramList[(512/Max[im])Flatten@im[[1;;-1;;ds,1;;-1;;ds]]];
	lmin = LMax[-GaussianFilter[histo[[2]], {6, 2}]];
	threshold = First@Extract[histo[[1]], Position[histo[[2]], Max[histo[[2]]]]];
	threshold = (Max[im]/512)First@Nearest[Select[lmin, # > threshold &], threshold];
	mask=UnitStep[im-Round[threshold]];
	ImageData@Closing[Opening[Image[mask,"Bit"], DiskMatrix[5]], DiskMatrix[5]]
	,-2,
	1-ImageData[RegionBinarize[ImageAdjust@Image@ArrayPad[im,1],{{1,1}},0.05]][[2;;-2,2;;-2]]
	,_,
	(*if -1 or manual specification*)
	threshold=If[thresholdIn==-1,FindThreshold[im]*0.25,thresholdIn];(*If not threshold is specified*)
	If[Length[Dimensions[im]]<3,(*In case the input image is RGB*)
		UnitStep[im-threshold],
		ImageData@Binarize@Image[UnitStep[im-threshold]]
	]
]
]

HeavyBlurringFast[im_,\[Sigma]_,downSampleSize_:8]:=ImageData@ImageResize[
		GaussianFilter[
			ImageResize[Image@im,Reverse@(Dimensions[im]/downSampleSize),Resampling-> "Gaussian"]
			,{4*\[Sigma]/downSampleSize,\[Sigma]/downSampleSize}
		]
		,Reverse@Dimensions[im],Resampling-> "Gaussian"
		]

IECorrection[im_,\[Sigma]In_:-1, speedUpFactor_:4]:=Module[{\[Sigma]},
	\[Sigma]=If[\[Sigma]In==-1,\[Sigma]=Min@(Dimensions[im]/16),\[Sigma]In];
	im/HeavyBlurringFast[im,\[Sigma]]]












Options[ESPreprocessImage] =
  {"FillBackground" -> True,
   "IECorrection" -> True,
   "Highpass" -> True};
ESPreprocessImage[im_, OptionsPattern[]] := Module[{imOut},
  imOut = im;
  If[OptionValue["FillBackground"],
   imOut = ESPaintBackground[imOut];
   ];
  If[OptionValue["IECorrection"],
   imOut = IECorrection[imOut,Length[imOut]/16];
   ];
  If[OptionValue["Highpass"],
   imOut = HighPass[imOut, Length[imOut]/16, 1];
   ];
  Return[imOut]
  ]

Options[PreprocessImage] =
  {"FillBackground" -> True,
   "IECorrection" -> True,
   "Highpass" -> True};
PreprocessImage[im_, OptionsPattern[]] := Module[{imOut, mask},
  imOut = im;
  If[OptionValue["FillBackground"],
   mask = CreateApertureMask[im,-1];
   imOut = InpaintBackground[imOut, mask];
   ];
  If[OptionValue["IECorrection"],
   imOut = IECorrection[imOut, Round[Length[im]/16]];
   ];
  If[OptionValue["Highpass"],
   imOut = HighPass[imOut, Round[Length[im]/16], 1];
   ];
  Return[imOut]
  ]
  
  
  
  
  
  
  
  
  
(*******************************************************)
(*******************************************End Private*)
End[]

(*******************************************************)
(*******************************************End Package*)
Protect["RetinalImageAnalysis`Preprocessing`*"]
EndPackage[]