(* Mathematica package *)

(*******************************************************)
(*****************************************Begin Package*)
BeginPackage["RetinalImageAnalysis`Tools`"]
Unprotect["RetinalImageAnalysis`Tools`*"]

ArgGMax::usage = "ArgGMax[list] returns the position of the global maximum in list."
ArgLMax::usage = "ArgLMax[list] returns the positions of the local maxima in list. 
By default constant boundary conditions are used, but this can be set to cyclic
with the options \"Cyclic\"->True."

(*******************************************************)
(*****************************************Begin Private*)
Begin["Private`"]


(*******************ArgGMax*)
ArgGMax1D = 
  Compile[{{list, _Real, 1}}, First@Position[list, Max[list]]];
ArgGMax2D = 
  Compile[{{list, _Real, 2}}, First@Position[list, Max[list]]];
ArgGMax3D = 
  Compile[{{list, _Real, 3}}, First@Position[list, Max[list]]];
ArgGMax4D = 
  Compile[{{list, _Real, 4}}, First@Position[list, Max[list]]];

ArgGMax[list_] := Switch[Length[Dimensions[list]]
  , 0,
  list
  , 1,
  First@ArgGMax1D[list]
  , 2,
  ArgGMax2D[list]
  , 3,
  ArgGMax3D[list]
  , 4,
  ArgGMax4D[list]
  , _,
  First@Position[list, Max[list]]
  ]
  
(*******************************ArgLMax*)

ArgLMax1D = 
 Compile[{{list, _Real, 1}, {cyclicQ, True | False}}, 
  Module[{higherThan},
   higherThan = ConstantArray[1, Length[list]];
   Do[higherThan = 
     UnitStep[list - RotateLeft[list, i] - $MachineEpsilon]*higherThan
    , {i, -1, 1, 2}];
   If[Not[cyclicQ],
    higherThan[[1]] = 0;
    higherThan[[-1]] = 0;
    ];
   Flatten@Position[higherThan, 1]
   ]
  ]

ArgLMax2D = 
 Compile[{{list, _Real, 2}, {cyclicQ, True | False}}, 
  Module[{higherThan},
   higherThan = ConstantArray[1, Dimensions[list]];
   Do[higherThan = 
     UnitStep[list - RotateLeft[list, {i, j}] - $MachineEpsilon]*
      higherThan
    , {i, -1, 1, 2}, {j, -1, 1, 2}];
   If[cyclicQ,
    Position[higherThan, 1]
    ,
    Position[higherThan[[2 ;; -2, 2 ;; -2]], 1] + 1
    ]
   ]
  ]


ArgLMax3D = 
 Compile[{{list, _Real, 3}, {cyclicQ, True | False}}, 
  Module[{higherThan},
   higherThan = 
    Table[1, {Dimensions[list][[1]]}, {Dimensions[list][[
       2]]}, {Dimensions[list][[3]]}];
   Do[higherThan = 
     UnitStep[list - RotateLeft[list, {i, j, k}] - $MachineEpsilon](**
    higherThan*)
    , {i, -1, 1, 2}, {j, -1, 1, 2}, {k, -1, 1, 2}];
   If[cyclicQ,
    Position[higherThan, 1]
    ,
    Position[higherThan[[2 ;; -2, 2 ;; -2, 2 ;; -2]], 1] + 1
    ]
   ]
  ]



ArgLMax4D = 
 Compile[{{list, _Real, 4}, {cyclicQ, True | False}}, 
  Module[{higherThan},
   higherThan = 
    Table[1, {Dimensions[list][[1]]}, {Dimensions[list][[
       2]]}, {Dimensions[list][[3]]}, {Dimensions[list][[4]]}];
   Do[higherThan = 
     UnitStep[
      list - RotateLeft[list, {i, j, k, l}] - $MachineEpsilon](**
    higherThan*)
    , {i, -1, 1, 2}, {j, -1, 1, 2}, {k, -1, 1, 2}, {l, -1, 1, 2}];
   If[cyclicQ,
    Position[higherThan, 1]
    ,
    Position[higherThan[[2 ;; -2, 2 ;; -2, 2 ;; -2, 2 ;; -2]], 1] + 1
    ]
   ]
  ]



Options[ArgLMax] = {"Cyclic" -> False};
ArgLMax[list_, OptionsPattern[]] := Switch[Length[Dimensions[list]]
  , 0,
  list
  , 1,
  ArgLMax1D[list, OptionValue["Cyclic"]]
  , 2,
  ArgLMax2D[list, OptionValue["Cyclic"]]
  , 3,
  ArgLMax3D[list, OptionValue["Cyclic"]]
  , 4,
  ArgLMax4D[list, OptionValue["Cyclic"]]
  , _,
  Print["Error!: Object dimensions to high. This function is only defined for 1D, 2D, 3D and 4D objects."];
  0
  ]





















(*******************************************************)
(*******************************************End Private*)
End[]

(*******************************************************)
(*******************************************End Package*)
Protect["RetinalImageAnalysis`Tools`*"]
EndPackage[]